import gevent
import gevent.monkey
import gevent.pywsgi
gevent.monkey.patch_all(thread=False)

import logging
import os
import signal
import socket
import sys
import urllib.parse

from django.conf import settings
from django.core.wsgi import get_wsgi_application
from django.utils.module_loading import import_string

from .. import watcher
from .. import manifest
from ..templatetags.asset import asset

from . import bottle
from .liveserver import LIVE_PORT
from .utils import IP_ANY

logger = logging.getLogger(__name__)



class Worker:
    RESTART_EXIT_CODE = 4
    RESTART_SIGNAL = signal.SIGHUP
    READY_SIGNAL = signal.SIGALRM

    LIVE_SCRIPT_TPL = "<script>window.livereload={'url':'{url}'}</script>"+asset('assets:javascript:livereload')


    @classmethod
    def _get_django_app(cls):
        app_path = getattr(settings, 'WSGI_APPLICATION')
        if app_path is None:
            return get_wsgi_application()
        else:
            return import_string(app_path)

    def _inject_live_script(self, app):
        def wrapper(environ, start_response):
            def wrapped_start_response(status, headerlist):
                headers = []
                for key, value in headerlist:
                    if key.lower() == 'Content-Length'.lower():
                        value = str(int(value)+len(self.live_script))
                    headers.append((key, value))
                return start_response(status, headers)
            injected = False
            for body_part in app(environ, wrapped_start_response):
                pos = body_part.find(b'</body>')
                if pos > 0:
                    yield body_part[:pos]
                    yield self.live_script
                    yield body_part[pos:]
                    injected = True
                else:
                    yield body_part
            if not injected:
                yield b' '*len(self.live_script)
        return wrapper

    @classmethod
    def spawn(cls, ip, port, live_url):
        logger.info("Spawning worker")
        args = [sys.executable] + ['-W%s' % o for o in sys.warnoptions] + sys.argv
        if sys.platform == "win32":
            args = ['"%s"' % arg for arg in args]
        new_environ = os.environ.copy()
        new_environ["_RUN_WORKER"] = 'true'
        new_environ["_WORKER_IP"] = ip
        new_environ["_WORKER_PORT"] = str(port)
        new_environ["_LIVE_URL"] = live_url

        worker_pid = os.spawnve(os.P_NOWAIT, sys.executable, args, new_environ)
        logger.info("Spawned worker PID {pid}".format(pid=worker_pid))
        return worker_pid

    @classmethod
    def is_worker(cls):
        return os.environ.get('_RUN_WORKER', 'false') == 'true'

    def __init__(self, ip=None, port=None, live_url=None):
        if ip is None:
            self.ip = os.environ.get('_WORKER_IP', IP_ANY)
        if port is None:
            self.port = int(os.environ.get('_WORKER_PORT', 8080))
        if live_url is None:
            self.live_url = os.environ.get('_LIVE_URL','ws://{ip}:{port}/'.format(ip=IP_ANY,port=LIVE_PORT))

        self.live_script = bytes(self.LIVE_SCRIPT_TPL.replace('{url}',self.live_url), encoding='ascii')

        # Determine whether we need a separate server
        # to serve static files

        static_url = urllib.parse.urlparse(settings.STATIC_URL)
        media_url = urllib.parse.urlparse(settings.MEDIA_URL)
        if static_url.hostname:
            static_ip = socket.gethostbyname(static_url.hostname)
        else:
            static_ip = IP_ANY
        if static_url.port:
            static_port = static_url.port
        else:
            static_port = self.port

        django_ip = self.ip
        django_port = self.port

        if (static_ip == IP_ANY or django_ip == IP_ANY or static_ip == django_ip) and (static_port == django_port):
            self._separate_static_server = False
        else:
            self._separate_static_server = True

        self.static_ip = static_ip
        self.static_port = static_port
        self.django_ip = django_ip
        self.django_port = django_port

        self.app = bottle.Bottle()

        # Add CORS headers
        @self.app.hook('after_request')
        def enable_cors():
            """
            You need to add some headers to each request.
            Don't use the wildcard '*' for Access-Control-Allow-Origin in production.
            """
            bottle.response.headers['Access-Control-Allow-Origin'] = '*'
            bottle.response.headers['Access-Control-Allow-Methods'] = 'PUT, GET, POST, DELETE, OPTIONS'
            bottle.response.headers['Access-Control-Allow-Headers'] = 'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token'

        # Serve static files
        @self.app.route(static_url.path+'<filename:path>', skip=[self._inject_live_script])
        def send_static(filename):
            return bottle.static_file(filename, root=settings.STATIC_ROOT)

        # Serve user-uploaded media files
        @self.app.route(media_url.path+'<filename:path>', skip=[self._inject_live_script])
        def send_static(filename):
            return bottle.static_file(filename, root=settings.MEDIA_ROOT)


    def signal_handler(self, sig_num, stack_frame):
        if sig_num == self.RESTART_SIGNAL:
            logger.info("Restarting worker.")
            exit(Worker.RESTART_EXIT_CODE)

    def run_forever(self):
        signal.signal(self.RESTART_SIGNAL, self.signal_handler)
        djapp = self._inject_live_script(self._get_django_app())

        if self._separate_static_server:
            logger.info("Starting separate static server on {ip}:{port}".format(ip=self.static_ip, port=self.static_port))
            gevent.spawn(gevent.pywsgi.WSGIServer((self.static_ip, self.static_port), self.app).serve_forever)
            server=gevent.pywsgi.WSGIServer((self.django_ip, self.django_port), djapp)
        else:
            @bottle.error(404)
            def error404(error):
                def start_response(status, headerlist):
                    bottle.response.status = int(status.split()[0])
                    for key, value in headerlist:
                        bottle.response.add_header(key, value)
                ret = djapp(bottle.request.environ, start_response)
                return ret
            server = gevent.pywsgi.WSGIServer((self.django_ip, self.django_port), self.app)
        logger.info("Starting django app on {ip}:{port}".format(ip=self.django_ip, port=self.django_port))
        server.start()
        os.kill(os.getppid(), self.READY_SIGNAL)
        server.serve_forever()






import errno
import inotify
import inotify.adapters
import inotify.calls
import inotify.constants
import logging
import os
import selectors

from django.conf import settings
from django.contrib.staticfiles.finders import get_finders
from django.contrib.staticfiles.storage import staticfiles_storage
from django.template.loaders.app_directories import get_app_template_dirs
from django.utils.autoreload import gen_filenames

FORMAT = '%(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)


def gen_template_filenames():
    bases = []
    tdir_names = ['templates']
    for tb in settings.TEMPLATES:
        bases.extend(tb.get('DIRS',[]))
        try:
            bases.extend(get_app_template_dirs(tb['OPTIONS']['app_dirname']))
        except:
            pass

    for template_dir in bases:
        for dir, dirnames, filenames in os.walk(template_dir):
            for filename in filenames:
                yield os.path.join(dir, filename)


class SelectorInotify(inotify.adapters.Inotify):
    def __init__(self, paths=[], block_duration_s=None):
        self._Inotify__block_duration = block_duration_s
        self._Inotify__watches = {}
        self._Inotify__watches_r = {}
        self._Inotify__buffer = b''
        self._Inotify__inotify_fd = inotify.calls.inotify_init()
        self._selector = selectors.DefaultSelector()

        self._selector.register(self._Inotify__inotify_fd, selectors.EVENT_READ)

        for path in paths:
            self.add_watch(path)

    def event_gen(self):
        while True:
            block_duration_s = self._Inotify__get_block_duration()

            # Poll, but manage signal-related errors.

            try:
                events = self._selector.select(block_duration_s)
            except IOError as e:
                if e.errno != errno.EINTR:
                    raise

                continue

            # Process events.

            for meta, event_type in events:
                # (fd) looks to always match the inotify FD.

                for (header, type_names, path, filename) \
                        in self._Inotify__handle_inotify_event(meta.fd, event_type):
                    yield (header, type_names, path, filename)

            yield None

class WatchError(BaseException):
    def __init__(self, ex):
        self._original_exception = ex
        if hasattr(ex, 'message'):
            msg = ex.message
        else:
            msg = str(ex)
        super().__init__(msg)

class Watcher:
    EVENT_MASK = (inotify.constants.IN_CLOSE_WRITE |
                  inotify.constants.IN_MODIFY |
                  inotify.constants.IN_DELETE_SELF |
                  inotify.constants.IN_MOVE_SELF |
                  inotify.constants.IN_CREATE |
                  inotify.constants.IN_MOVED_FROM |
                  inotify.constants.IN_MOVED_TO)
    STATIC = 0
    APP = 1

    def __init__(self):
        self.inotify = SelectorInotify()
        self.target_storage = staticfiles_storage
        self.source_storages = set()
        self.watch_map = {}
        self._app_files = []
        self._app_templates = []
        self.reload()

    def reload(self):
        self.unwatch()
        for finder in get_finders():
            if hasattr(finder, 'storages'):
                for s in finder.storages.values():
                    if hasattr(s,'watch_sources'):
                        self.source_storages.add(s)
            elif hasattr(finder, 'storage') and hasattr(finder.storage,'watch_sources'):
                self.source_storages.add(finder.storage)

        for storage in self.source_storages:
            storage.watch_sources(self.inotify, mask=Watcher.EVENT_MASK)

        self._app_files = list(gen_filenames())
        self._app_templates = list(gen_template_filenames())
        for f in self._app_files + self._app_templates:
            self.inotify.add_watch(bytes(os.path.dirname(f),encoding='utf-8'), mask=Watcher.EVENT_MASK)

    def reload_storage(self, storage):
        storage.load_manifest()
        storage.collect(self.target_storage)
        if storage in self.source_storages:
            storage.unwatch_sources(self.inotify)
        storage.watch_sources(self.inotify, mask=Watcher.EVENT_MASK)

    def unwatch(self):
        for storage in self.source_storages:
            storage.unwatch_sources(self.inotify)
        for f in self._app_files + self._app_templates:
            self.inotify.remove_watch(bytes(f,encoding='utf-8'))

    def changes(self):
        try:
            for event in self.inotify.event_gen():
                if event is not None:
                    header, type_names, watch_path, filename = event
                    changed_src_path = os.path.join(watch_path.decode('utf-8'), filename.decode('utf-8')).rstrip('/')
                    if changed_src_path in self._app_files:
                        yield (changed_src_path, Watcher.APP)
                    elif changed_src_path in self._app_templates:
                        yield (changed_src_path, Watcher.STATIC)
                    for s in self.source_storages:
                        try:
                            fname = s.get_changed_file(event)
                            if fname is s.C_MANIFEST:
                                self.reload_storage(s)
                                yield (s._manifest_path, Watcher.STATIC)
                                break
                            elif fname is not None:
                                s.compile_file(fname, self.target_storage)
                                yield (s.prefixed_path(fname), Watcher.STATIC)
                                break
                        except Exception as ex:
                            logger.warn("Error when compiling changed path '"+s.prefixed_path(str(fname))+"':"+str(ex))

        except Exception as ex:
            raise WatchError(ex)
        finally:
            self.unwatch()







from assets.storage import ManifestTree

def test_init():
    root = ManifestTree(path='test/subdir/A')
    assert list(root.list_dirs()) == ['subdir']
    assert list(root['subdir'].list_dirs()) == ['A']
    assert list(root['subdir']['A'].list_dirs()) == []


def test_frompath():
    root, leaf = ManifestTree.from_path(path='test/subdir/A')
    assert list(root.list_dirs()) == ['subdir']
    assert list(root['subdir'].list_dirs()) == ['A']
    assert list(root['subdir']['A'].list_dirs()) == []
    assert leaf.name == 'A'
    assert list(leaf.list_dirs()) == []

def test_iterfiles():
    root = ManifestTree()
    sub = ManifestTree('sub')
    sub._files['testC'] = 'C'
    root.append_tree(sub)
    root._files['testA'] = 'A'
    root._files['testB'] = 'B'
    root.append_tree(sub)
    assert set(root.iter_files()) == set([('testA','A'),('testB','B'),('sub/testC','C')])



import base64
import datetime
import functools
import hashlib
import json
import logging
import os
import re
import pathlib

from django.core.files.base import ContentFile
from django.contrib.staticfiles import storage

from .filters import get_filter_by_name

FORMAT = '%(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger(__name__)

def path_strip(path, left=0, right=0):
    if right == 0:
        right = len(path.parts)
    else:
        right = -right
    ret = pathlib.Path()
    for part in path.parts[left:right]:
        ret /= part
    return ret

def list_files(path, path_pattern='.*'):
    ret = []
    if isinstance(path_pattern, str):
        path_pattern = re.compile(path_pattern)
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    for item in path.iterdir():
        if item.is_file() and path_pattern.search(str(item)):
            ret.append(item)
        elif item.is_dir():
            ret += list_files(item, path_pattern=path_pattern)
    return ret


class ManifestTree:
    """
        Represents the target directory layout as constructed from a manifest file.
    """

    @classmethod
    def from_path(cls, path):
        """
            Given a path, constructs a tree from `path` (i.e. a tree where
            each node has at most one directory) and returns the root of the
            tree and the unique leaf node of the tree.
        """
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        # Construct the tree
        root = node = ManifestTree(path)

        # Find the (unique) leaf node
        while len(node._dirs) > 0:
            node = list(node._dirs.values())[0]

        return root, node

    def __init__(self, path='', data=None):
        """
            The constructor creates a tree struture corresponding to `path`.
            If `path` is emtpy or has just one component, the tree will
            consist only of the root node whose name will be `path.name`
            and whose `_data` attribute will contain `data`. If `path` has more
            components, a corresponding tree structure is built up (each node
            except for the leaf having precisely one successor) and the
            unique leaf node of the tree will have its `_data` attribute set to `data`.
        """
        if data is None:
            data = {}

        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        self._files = {}

        # Used for storing subdirectories (instances of ManifestTree)
        # keyed by subdirectory name
        self._dirs = {}

        if len(path.parts) > 1:
            self._name = path.parts[0]
            self._dirs[path.parts[1]] = ManifestTree(path_strip(path,left=1), data)
        else:
            self._data = data
            self._name = path.name

    def append_tree(self, tree):
        """
            Inserts `tree` as a subdirectory whose name will be `tree.name`
        """
        self._dirs[tree.name] = tree

    def __getitem__(self, key):
        """
            Returns the data attribute if `key` is a file, otherwise
            returns the subdirectory named `key` (an instance of ManifestTree)
        """
        if isinstance(key, pathlib.Path):
            key = key.name
        if key in self._files:
            return self._files[key]
        return self._dirs[key]

    def __setitem__(self, key, value):
        """
            If value is a ManifestTree, the method is equivalent to `append_tree(value)`,
            otherwise inserts the value into the `files` dictionary.
        """
        if isinstance(key, pathlib.Path):
            key = key.name
        if isinstance(value, ManifestTree):
            assert value.name == key
            self.append_tree(value)
        self._files[key] = value

    def __contains__(self, key):
        """
            Returns true if the current node has a file or directory whose
            name is `key`.
        """
        if isinstance(key, pathlib.Path):
            key = key.name
        return key in self._files or key in self._dirs

    def load_dir(self, path, file_pattern, dir_pattern, target_pattern=None, default_data=None):
        """
            Grafts a FileSystem directory tree starting at `path` onto the
            current node associating with each file metadata given by `default_data`.
            Only includes paths matching `file_pattern` (which must be an object
            implementing a match method) and only recurses into directories
            matching the `dir_pattern` (which must be an object implementing a
            match method).
        """
        if default_data is None:
            default_data = {}
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        for item in path.iterdir():
            if item.is_file() and file_pattern.match(item.name):
                source = str(item)
                if target_pattern:
                    item = pathlib.Path(file_pattern.sub(target_pattern, str(item)))
                self._files[item.name] = {
                    'name': item.stem,
                    'ext': item.suffix,
                    'sources': [source],
                    'filters': []
                }
                self._files[item.name].update(default_data)
            elif item.is_dir() and dir_pattern.match(item.name):
                dtree = ManifestTree(item.name, data=default_data)
                dtree.load_dir(item, file_pattern, dir_pattern, target_pattern, default_data)
                self._dirs[item.name] = dtree

    def find(self, path):
        """
            Returns the metadata associated with `path`, if `path`
            exists in the manifest tree.
        """
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        if len(path.parts) == 0:
            return self._data
        elif len(path.parts) == 1:
            return self._files[path.name]
        else:
            dtree = self._dirs[path.parts[0]]
            return dtree.find(path_strip(path,left=1))

    def listdir(self, path):
        """
            Returns a pair consisting of the list of subdirectories and the list of files
            of the directory `path` in the ManifestTree
        """
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)

        if len(path.parts) == 0:
            return self.list_dirs(), self.list_files()

        return self._dirs[path.parts[0]].listdir(path_strip(path,left=1))

    def list_files(self):
        """
            Lists the files contained in the node.
        """
        return self._files.keys()

    def list_dirs(self):
        """
            Lists the directories contained in the node.
        """
        return self._dirs.keys()

    def iter_files(self):
        for fname, data in self._files.items():
            yield (fname, data)

        for dirname, dirtree in self._dirs.items():
            for fname, data in dirtree.iter_files():
                yield (os.path.join(dirname, fname), data)

    @property
    def name(self):
        return self._name

    @property
    def data(self):
        return self._data

    def pretty_str(self, indent=0):
        istr = " "*indent*4
        files="".join([istr+" "+
                       fname+
                       "("+",".join(data.get('sources',[]))+")\n"
                       for fname, data in self._files.items()])
        dirs = "".join([istr+"-"+name+"\n"+d.pretty_str(indent+1) for name, d in self._dirs.items()])
        ret = []
        if len(files) > 0:
            ret.append(files)
        if len(dirs) > 0:
            ret.append(dirs)
        if len(ret) > 0:
            return "\n".join(ret)+"\n"
        else:
            return "(EMPTY)\n"

    def __str__(self):
        return self.pretty_str()


def conditionally_overrides(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._manifest is None:
            return getattr(super(self.__class__, self), method.__name__)(*args, **kwargs)
        else:
            return method(self, *args, **kwargs)
    return wrapper


class ManifestStorage(storage.FileSystemStorage):
    """
    Manifest filesystem storage

    If the static directory doesn't contain a `manifest.json`
    file, this storage behaves exactly like :class:`FileSystemStorage`.
    Otherwise the `manifest.json` file has the following format (example):

    {
        # The top-level keys denote scopes/asset types
        # Use them as prefixes to the asset tag
        # e.g. {% asset 'app:javascript:bootstrap' %}

        'javascript': {
            'target_dir':'js',  # The name of the target directory where assets of this type are stored
            'target_ext':'.js', # The extension appended to all assets of this type
            'items': {          # A list of assets
                'bootstrap': {  # Asset name (used in the asset tag)
                    'sources': [# The asset is constructed by concatenating these files and
                                # optionally passing them through specified filters
                        'sass-bootstrap/js/tooltip.js',
                        'sass-bootstrap/js/popover.js',
                        'sass-bootstrap/js/tab.js',
                        'bootstrap-hover-dropdown/bootstrap-hover-dropdown.js'
                    ],
                    'filters':['jsmin'] # The filters to pass the source files through
                }
            }
        },

        'stylesheets': {
            'target_dir':'css',
            'target_ext':'css',
            'items': {
                'logika': {
                    'sources':['sass/logika.scss'],
                    'filters':['sass','cssmin']
                }
            }
        }
    }

    """

    C_MANIFEST = 0
    _CHECKSUM_NAME = '__checksums__'

    def __init__(self, *args,  app_name="", **kwargs):
        super().__init__(*args, **kwargs)
        self._manifest = None
        self._checksums = {}
        if super().exists('manifest.json'):
            self._app_name = app_name
            self.load_manifest()
            logging.info("Loaded: " + self.path('manifest.json'))

    @property
    def _manifest_path(self):
        """The location of `manifest.json`"""
        if super().exists('manifest.json'):
            return self.path('manifest.json')
        else:
            return None

    @property
    def checksum_name(self):
        if hasattr(self, '_app_name'):
            return os.path.join(self._app_name, self._CHECKSUM_NAME)
        else:
            return self._CHECKSUM_NAME

    def load_manifest(self):
        with open(self._manifest_path, 'r') as IN:
            try:
                manifest = json.load(IN)
            except Exception as ex:
                raise Exception(str(ex)+" (when loading "+self._manifest_path+")")
        self._manifest_data = manifest
        self._manifest = append_to = ManifestTree()

        if len(self._app_name) > 0:
            append_to = ManifestTree(self._app_name)
            self._manifest.append_tree(append_to)

        # Iterate over asset types
        for tp in manifest.keys():
            target_dir = pathlib.Path(manifest[tp]['target_dir'])
            target_ext = manifest[tp].get('target_ext', '')
            troot, tleaf = ManifestTree.from_path(target_dir)
            append_to.append_tree(troot)
            if 'items' in manifest[tp]:
                for name, meta in manifest[tp]['items'].items():
                    tleaf[name+target_ext] = {
                        'scope': tp,
                        'name': name,
                        'ext': target_ext,
                        'sources': meta['sources'],
                        'filters': meta.get('filters', []),
                        'watch': meta.get('watch', [])
                    }
            elif 'copy' in manifest[tp]:
                file_pattern = re.compile(manifest[tp]['copy'].get('pattern', '.*'))
                dir_pattern = re.compile(manifest[tp]['copy'].get('dir_pattern', '.*'))
                target_pattern = manifest[tp]['copy'].get('target_pattern', None)
                for source in manifest[tp]['copy']['sources']:
                    tleaf.load_dir(self.path(source), file_pattern=file_pattern, dir_pattern=dir_pattern, target_pattern=target_pattern, default_data={'scope': tp, 'filters': manifest[tp]['copy'].get('filters', [])})

    def prefixed_path(self, fname):
        if getattr(self, 'prefix', None):
            return os.path.join(self.prefix, fname)
        return fname

    def compile_file(self, fname, target_storage):
        """Compiles the single static file :param:`fname` and saves it to :param:`target_storage`"""
        prefixed_path = self.prefixed_path(fname)
        with self.open(fname) as SRC:
            if fname == self.checksum_name:
                logger.info("Collecting checksums: " +prefixed_path)
            target_storage.delete(prefixed_path)
            target_storage.save(prefixed_path, SRC)

        if not fname == self.checksum_name:
            self.compile_file(self.checksum_name, target_storage)

    def collect(self, target_storage):
        """Compiles all static files and saves them to :param:`target_storage`"""
        for fname, _ in self.iter_files():
            self.compile_file(fname, target_storage)

    @conditionally_overrides
    def _open(self, name, mode):
        if mode != 'r' and mode != 'rb':
            raise NotImplementedError("Manifest Storage only supports opening in 'r' mode, got '"+mode+"'.")
        if name == self.checksum_name:
            return ContentFile(json.dumps(self._checksums, sort_keys=True, indent=4), name=name)
        item = self._manifest.find(name)
        data = b''
        cwd = self.path('')
        if item.get('filters', []):
            logger.info("Compiling '"+str(name)+"'")
            logger.info("  sources: "+str(item['sources']))
            logger.info("  filters: "+str(item.get('filters', [])))
        else:
            logger.info("Copying '"+str(name)+"'")
        for src in item['sources']:
            cwd, _ = os.path.split(self.path(src))
            data = data+open(self.path(src), 'rb').read()
        config = item.get('config', {})
        for filt in item.get('filters', []):
            f = get_filter_by_name(filt)
            data = f(data, cwd=cwd, config=config)
        if isinstance(data, str):
            data = bytes(data, encoding='utf-8')
        self._checksums[name] = str(base64.b64encode(hashlib.sha256(data).digest()), encoding='ascii')
        logger.info("  checksum: "+self._checksums[name])
        return ContentFile(data, name=name)

    @conditionally_overrides
    def delete(self, path):
        raise NotImplementedError("Manifest Storage does not support deleting.")

    @conditionally_overrides
    def listdir(self, path):
        return self._manifest.listdir(path)

    @conditionally_overrides
    def exists(self, name):
        if name == self.checksum_name:
            return True
        try:
            item = self._manifest.find(name)
            return True
        except:
            return False

    @conditionally_overrides
    def size(self, name):
        raise NotImplementedError("Manifest Storage does not support the size call")

    @conditionally_overrides
    def get_accessed_time(self, name):
        if name == self.checksum_name:
            return datetime.now()
        item = self._manifest.find(name)
        latest_stamp = 0
        for s in item['sources']:
            stamp = os.path.getatime(self.path(s))
            if stamp > latest_stamp:
                latest_stamp = stamp
        return datetime.fromtimestamp(latest_stamp)

    @conditionally_overrides
    def get_modified_time(self, name):
        if name == self.checksum_name:
            return datetime.now()
        item = self._manifest.find(name)
        latest_stamp = 0
        for s in item['sources']:
            stamp = os.path.getmtime(self.path(s))
            if stamp > latest_stamp:
                latest_stamp = stamp
        return datetime.fromtimestamp(latest_stamp)

    @conditionally_overrides
    def get_created_time(self, name):
        raise NotImplementedError("Create time not available")

    def iter_files(self, path=''):
        if self._manifest is None:
            dirs, files = self.listdir(path)
            for f in files:
                yield (os.path.join(path,f), {'sources':[os.path.join(path,f)]})
            for d in dirs:
                try:
                    for f, data in self.iter_files(path=os.path.join(path,d)):
                        yield (f, data)
                except:
                    logger.error("Error listing directory "+os.path.join(path,d) + "("+self._app_name+")")
        else:
            for f, data in self._manifest.iter_files():
                yield f, data
            yield self.checksum_name, None

    def watch_sources(self, inotify, mask):
        if self._manifest:
            logger.info("Registering watches for "+self.path(''))
            inotify.add_watch(bytes(self._manifest_path, encoding='utf-8'), mask=mask)
        self._watch_map = {}
        for fname, data in self.iter_files():
            if fname == self.checksum_name:
                continue
            source_files = set([self.path(src) for src in data['sources']])
            source_dirs = set([os.path.dirname(f) for f in source_files])
            if 'watch' in data:
                for watch_pattern in data['watch']:
                    path = pathlib.Path('.')
                    for p in list_files(path, watch_pattern):
                        if p.is_file():
                            source_dirs.add(str(p.parent.absolute()))
                            source_files.add(str(p.absolute()))
            for src in source_dirs:
                inotify.add_watch(bytes(src,encoding='utf-8'),mask=mask)
            for src in source_files:
                self._watch_map[src] = fname

    def unwatch_sources(self, inotify):
        for p in self._watch_map.keys():
            inotify.remove_watch(bytes(p,encoding='utf-8'))
        if self._manifest:
            inotify.remove_watch(bytes(self._manifest_path, encoding='utf-8'))

    def get_changed_file(self, inotify_event):
        """
            Returns the name of the file (as given by iter_files) which is affected
            by the change reported by :param:`inotify_event`.
        """
        header, type_names, watch_path, filename = inotify_event
        changed_src_path = os.path.join(watch_path.decode('utf-8'), filename.decode('utf-8')).rstrip('/')
        if changed_src_path in self._watch_map:
            target = self._watch_map[changed_src_path]
            return target
        elif changed_src_path == self._manifest_path:
            return self.C_MANIFEST




import gevent
import gevent.monkey
gevent.monkey.patch_all(thread=False)

import logging
import os
import signal

from django.conf import settings

from .. import watcher

from . import bottle
from .liveserver import LiveServer, LIVE_PORT
from .utils import IP_ANY, CleanChildProcesses, pid_exists, with_lock
from .wsgi_worker import Worker

logger = logging.getLogger(__name__)


class Supervisor:
    PID_NAME = 'livereload.pid'
    RELOAD_SIGNAL = signal.SIGUSR1
    RESTART_SIGNAL = signal.SIGHUP

    WORKER_DEAD = 0
    WORKER_STARTING = 1
    WORKER_READY = 2
    WORKER_CLOSING = 3

    @classmethod
    def pid_path(cls):
        # FIXME: Eventually figure out how to get the path of the main app.py
        #        so that the pid file is created there.
        return cls.PID_NAME

    @classmethod
    def save_pid(cls):
        pid = str(os.getpid())
        pid_path = cls.pid_path()
        with open(pid_path, 'w', encoding='utf-8') as pid_file:
            logger.info("Writing supervisor PID " + pid + " to " + pid_path)
            pid_file.write(pid)

    @classmethod
    def kill(cls, sig_num):
        pid_path = cls.pid_path()
        if os.path.exists(pid_path):
            with open(pid_path, 'r', encoding='utf-8') as pid_file:
                try:
                    server_pid = int(pid_file.read())
                    os.kill(server_pid, sig_num)
                    return True
                except:
                    pass
        return False

    def __init__(self, django_ip=IP_ANY, django_port=8080, live_port=LIVE_PORT, live_ip=None):
        logger.info("Initializing Supervisor")
        self._live_server = LiveServer(ip=live_ip or django_ip, port=live_port)
        self.dj_ip = django_ip
        self.dj_port = django_port
        self._worker_pid = None
        self._worker_status = self.WORKER_DEAD
        self._worker_lock = gevent.lock.BoundedSemaphore(1)

        # Kill any previously running supervisor
        killed_old = self.kill(signal.SIGKILL)
        if killed_old:
            logger.info("Killed previously running supervisor")
        self.save_pid()

        # Install signal handlers
        signal.signal(self.RELOAD_SIGNAL, self.signal_handler) # Reload client
        signal.signal(self.RESTART_SIGNAL, self.signal_handler)  # Restart child server
        signal.signal(Worker.READY_SIGNAL, self.signal_handler) # Child server ready

    def signal_handler(self, sig_num, stack_frame):
        if sig_num == self.RELOAD_SIGNAL:
            logger.info("Received signal {sig} forcing client reload.".format(sig=sig_num))
            self._live_server.reload()
        elif sig_num == self.RESTART_SIGNAL:
            logger.info("Received signal {sig} restarting server.".format(sig=sig_num))
            self.restart_worker()
        elif sig_num == Worker.READY_SIGNAL:
            logger.info("Received signal {sig} worker ready.".format(sig=sig_num))
            self._worker_status = self.WORKER_READY
            self._live_server.status('ready')
            self._live_server.reload()

    def _path2url(self, path):
        return settings.STATIC_URL+'/'+path

    def _watcher(self):
        logger.info("Staring watcher")
        w = watcher.Watcher()
        for path, file_type in w.changes():
            if path is not None:
                if file_type == watcher.Watcher.STATIC:
                    logger.info("Reloading client because '{path}' changed".format(path=path))
                    self._live_server.reload(self._path2url(path))
                else:
                    logger.info("Restarting server because '{path}' changed".format(path=path))
                    self.restart_worker()

    def _spawn_worker(self):
        with self._worker_lock:
            self._live_server.status('starting')
            self.kill_worker(Worker.RESTART_SIGNAL)

            # If the previous worker didn't exit by now, kill it forcefully
            self.kill_worker(signal.SIGKILL)

            self._worker_pid = Worker.spawn(ip=self.dj_ip, port=self.dj_port, live_url=self._live_server.url)
            self._worker_status = self.WORKER_STARTING

    def restart_worker(self):
        if self._worker_lock.locked():
            return False
        self._spawn_worker()


    def kill_worker(self, signum):
        if pid_exists(self._worker_pid):
            try:
                logger.info("Sending signal {sig} to worker PID {pid}".format(pid=self._worker_pid, sig=signum))
                os.kill(self._worker_pid, signum)
            except:
                pass

    def _watch_children(self):
        while True:
            pid, exit_status = os.waitpid(-1, options=0)
            exit_code = os.WEXITSTATUS(exit_status)
            if not self._worker_lock.locked() and pid == self._worker_pid:
                if exit_code != Worker.RESTART_EXIT_CODE:
                    logger.info("Worker died with unexpected exit code "+str(exit_code))
                    logger.info("Waiting for source change/signal to restart")
                self._live_server.status('dead')
                self._worker_status = self.WORKER_DEAD

    def run_forever(self):
        gevent.spawn(self._watcher)
        gevent.spawn(self._live_server.serve_forever)

        with CleanChildProcesses():
            self._spawn_worker()
            self._watch_children()






import json
import os
import re
import subprocess
import sys
from tempfile import NamedTemporaryFile, gettempdir
from pathlib import Path

from . import register_filter
from .utils import pipe

try:
    import shutil

    if shutil.which('webpack') is not None:
        @register_filter('babel')
        def babel_filter(input_bytes, cwd, config={}):
            """
            A filter for bundling javascript with webpack using the babel loader. Provide
            babel options under the "config.babel" key of a given item in manifest.json
            """

            # Create temporary files
            #    temp_in contains the source to transform
            #    temp_out will contain the transformed source
            temp_in = NamedTemporaryFile(suffix='.js', dir=cwd)
            temp_out = NamedTemporaryFile(suffix='.js')
            temp_cfg = NamedTemporaryFile(suffix='.js', dir=cwd)

            # Construct the configuration
            babel_options = {}
            if hasattr(babel_filter, "_config"):
                babel_options.update(babel_filter._config)
            babel_options.update(config.get('babel', {}))

            webpack_config = {
                "entry": temp_in.name,
                "output": {
                    "path": gettempdir(),
                    "filename": Path(temp_out.name).name
                },
                "module": {
                    "rules": [{
                        "test": "/\.js$/",
                        "exclude": "/(node_modules|bower_components)/",
                        "use": {
                            "loader": "babel-loader",
                            "options": babel_options,
                        }
                    }]
                }
            }

            # Write the source & config to temporary files
            temp_in.write(input_bytes)
            temp_in.flush()
            temp_cfg.write(bytes("const config="+json.dumps(webpack_config, indent=4)+";module.exports=config;", encoding='utf-8'))
            temp_cfg.flush()

            # Run webpack
            try:
                p = subprocess.run(["webpack", "--config", temp_cfg.name], stdout=sys.stdout, stderr=sys.stderr, cwd=cwd, check=True)
            except:
                print("Error running webpack")
                print("Config:", Path(temp_cfg.name).read_text())
                raise

            # Read the output
            temp_out.seek(0)
            ret = temp_out.read()

            # Cleanup
            temp_out.close()
            temp_in.close()
            temp_cfg.close()

            return ret

    else:
        @register_filter('babel')
        def babel_filter(bytes, cwd, config={}):
            """
            A stub for babel converter (the stub does nothing)
            """
            return bytes
except:
    @register_filter('babel')
    def babel_filter(bytes, cwd, config={}):
        """
        The webpack program not found, this is just a stub. To install
        run

        $ npm -i -g webpack

        """
        return bytes






import os
from collections import OrderedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from django.apps import apps
from django.contrib.staticfiles import finders

from .storage import ManifestStorage


class ManifestAppDirsFinder(finders.AppDirectoriesFinder):
    storage_class = ManifestStorage

    def __init__(self, app_names=None, *args, **kwargs):
        # The list of apps that are handled
        self.apps = []
        # Mapping of app names to storage instances
        self.storages = OrderedDict()
        app_configs = apps.get_app_configs()
        if app_names:
            app_names = set(app_names)
            app_configs = [ac for ac in app_configs if ac.name in app_names]
        for app_config in app_configs:
            app_storage = self.storage_class(
                os.path.join(app_config.path, self.source_dir), app_name=app_config.name)
            if os.path.isdir(app_storage.location):
                self.storages[app_config.name] = app_storage
                if app_config.name not in self.apps:
                    self.apps.append(app_config.name)


class ManifestFinder(finders.FileSystemFinder):
    def __init__(self, app_names=None, *args, **kwargs):
        # List of locations with static files
        self.locations = []
        # Maps dir paths to an appropriate storage instance
        self.storages = OrderedDict()
        if not isinstance(settings.STATICFILES_DIRS, (list, tuple)):
            raise ImproperlyConfigured(
                "Your STATICFILES_DIRS setting is not a tuple or list; "
                "perhaps you forgot a trailing comma?")
        for root in settings.STATICFILES_DIRS:
            if isinstance(root, (list, tuple)):
                prefix, root = root
            else:
                prefix = ''
            if settings.STATIC_ROOT and os.path.abspath(settings.STATIC_ROOT) == os.path.abspath(root):
                raise ImproperlyConfigured(
                    "The STATICFILES_DIRS setting should "
                    "not contain the STATIC_ROOT setting")
            if (prefix, root) not in self.locations:
                self.locations.append((prefix, root))
        for prefix, root in self.locations:
            filesystem_storage = ManifestStorage(location=root)
            filesystem_storage.prefix = prefix
            self.storages[root] = filesystem_storage




from . import register_filter
from .utils import pipe

try:
    import shutil

    if shutil.which('clojure-compiler') is not None:
        @register_filter('clojure')
        def clojure_filter(bytes, cwd, config={}):
            """
            A JS minifier using Google's clojure compiler
            """
            return pipe('clojure-compiler', bytes, cwd=cwd)
    else:
        @register_filter('clojure')
        def clojure_filter(bytes, cwd, config={}):
            """
            A stub for Google's clojure compiler (the stub does nothing)
            """
            return bytes
except:
    @register_filter('clojure')
    def clojure_filter(bytes, cwd, config={}):
        """
        A stub for Google's clojure compiler (the stub does nothing)
        """
        return bytes



from django.core.management.base import BaseCommand, CommandError

from ..supervisor import Supervisor

class Command(BaseCommand):
    help = 'Reload all open browser windoes.'

    def handle(self, *args, **options):
        Supervisor.kill(Supervisor.RELOAD_SIGNAL)



import os
import json
import re

from django.apps import apps
from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.exceptions import ImproperlyConfigured
from django.contrib.staticfiles.templatetags.staticfiles import static

from .storage import ManifestStorage


class Manifest:
    ASSET_LIST = {}

    source_dir = 'static'
    loaded = False
    checksums = {}

    @classmethod
    def to_id(cls, prefix, scope, name):
        """
            Creates a bundle id out of its components, where :param:`prefix` is intended to be
            :param:`scope` the asset type (e.g. javascript, ...) and :param:`name` the name
            of the asset."""
        if prefix is None:
            return ':'.join([scope, name])
        else:
            return ':'.join([prefix, scope, name])

    @classmethod
    def from_id(cls, id):
        """
            Splits a bundle id into its components prefix, scope, name, where :param:`prefix`
            is intended to be :param:`scope` the asset type (e.g. javascript, ...) and
            :param:`name` the name of the asset."""
        parts = id.split(':')
        if len(parts) == 3:
            return parts[0], parts[1], parts[2]
        else:
            return None, parts[0], parts[1]

    @classmethod
    def _tpath(cls, prefix, scope_data, name):
        if prefix is None:
            return os.path.join(scope_data.get('target_dir',''), name) + scope_data.get('target_ext', '')
        else:
            return os.path.join(prefix, scope_data.get('target_dir',''), name) + scope_data.get('target_ext', '')

    @classmethod
    def _match(self, key, id):
        parts = id.split(':')
        pattern = parts[-1]
        stem = id[:-len(pattern)]

        # The prefix:scope: must match exactly
        if not key.startswith(stem):
            return False

        # Now we deal with different patterns

        # Case ``start*``
        if key.startswith(pattern.rstrip('*')):
            return True

        # Case ``*middle*``
        if pattern[0] == '*' and pattern[-1] == '*':
            return pattern.strip('*') in key

        return re.match(pattern, key)

    @classmethod
    def _load_manifest(cls, manifest, prefix=None):
        try:
            if not prefix in cls.checksums:
                cls.checksums[prefix] = {}
                print("Loading checksums from:", cls._tpath(prefix, {}, ManifestStorage._CHECKSUM_NAME))
            with staticfiles_storage.open(cls._tpath(prefix, {}, ManifestStorage._CHECKSUM_NAME)) as CHECKSUMS:
                cls.checksums[prefix].update(json.loads(str(CHECKSUMS.read(), encoding='utf-8')))
        except Exception as ex:
            print("Failed to load checksums", str(ex))
            pass
        for scope_name, scope_data in manifest.items():
            if 'items' in scope_data:
                for item_name, item_data in scope_data['items'].items():
                    key = cls.to_id(prefix, scope_name, item_name)
                    val = cls._tpath(prefix, scope_data, item_name)
                    item_data['checksum'] = cls.checksums[prefix].get(val, '')
                    cls.ASSET_LIST[key] = {'url': static(val), 'data': item_data}
            elif 'copy' in scope_data:
                key = cls.to_id(prefix, scope_name, '*')
                val = cls._tpath(prefix, scope_data, '')
                cls.ASSET_LIST[key] = {'url': static(val), 'data': {}, '__checksum_base__': val}

    @classmethod
    def load(cls):
        locations = []
        app_configs = apps.get_app_configs()
        for app_config in app_configs:
            manifest_path = os.path.join(app_config.path, cls.source_dir, 'manifest.json')
            if os.path.exists(manifest_path):
                manifest = json.load(open(manifest_path, 'r'))
                cls._load_manifest(manifest, app_config.name)
        for root in settings.STATICFILES_DIRS:
            if isinstance(root, (list, tuple)):
                prefix, root = root
            else:
                prefix = ''
            if settings.STATIC_ROOT and os.path.abspath(settings.STATIC_ROOT) == os.path.abspath(root):
                raise ImproperlyConfigured(
                    "The STATICFILES_DIRS setting should "
                    "not contain the STATIC_ROOT setting")
            if (prefix, root) not in locations:
                locations.append((prefix, root))
        for prefix, root in locations:
            manifest_path = os.path.join(root, 'manifest.json')
            if os.path.exists(manifest_path):
                manifest = json.load(open(manifest_path, 'r'))
                cls._load_manifest(manifest, None)

    @classmethod
    def find_assets(cls, bundle_id):
        """
            Returns a list of url, data pairs describing
            the static location of the files in bundle `bundle_id`
            and the data describing the file (e.g. checksum, ...)
        """
        if not cls.loaded:
            cls.load()
        prefix, scope, name = cls.from_id(bundle_id)
        assets = []
        if bundle_id in cls.ASSET_LIST:
            url = cls.ASSET_LIST[bundle_id]['url']
            data = cls.ASSET_LIST[bundle_id]['data']
            assets.append((url, data))
        else:
            key = cls.to_id(prefix, scope, '*')
            if key in cls.ASSET_LIST:
                url = cls.ASSET_LIST[key]['url']+'/'+name
                data = cls.ASSET_LIST[key]['data']
                data['checksum'] = cls.checksums[prefix].get(cls.ASSET_LIST[key].get('__checksum_base__','')+'/'+name, '')
                assets.append((url, data))
            else:
                for (key, val) in cls.ASSET_LIST.items():
                    if cls._match(key, bundle_id):
                        assets.append((val['url'], val['data']))
        return assets



from . import register_filter

try:
    from cssmin import cssmin

    @register_filter('cssmin')
    def cssmin_filter(bytes_input, cwd, config={}):
        """
            A CSS minifier based on an (unmaintained) Python port of the YUI cssmin program
        """
        return cssmin(bytes_input)
except:
    @register_filter('cssmin')
    def cssmin_filter(bytes, cwd, config={}):
        """
        A stub for the cssmin filter (the stub does nothing)
        """
        return bytes




import logging

from django.core.management.base import BaseCommand, CommandError
from ...filters import get_all_filters

FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'List available asset filters'

    def handle(self, *args, **options):
        for name, f in get_all_filters():
            print(name, f.__doc__)






from django.core.management.base import BaseCommand, CommandError

from ..supervisor import Supervisor
from ..wsgi_worker import Worker

class Command(BaseCommand):
    help = 'Serve static files using a CORS-enabled http.server'

    def handle(self, *args, **options):
        if Worker.is_worker():
            agent = Worker()
        else:
            agent = Supervisor()
        agent.run_forever()







from collections import defaultdict

from django import template
from django.template import Context, Template
from django.utils.safestring import mark_safe


# If we have jinja2 templates
try:
    from django_jinja import library
except:
    class library:
        global_function = lambda x: x
        filter = lambda x: x

from ..manifest import Manifest

register = template.Library()


class SimpleTemplate:
    def __init__(self, code):
        self.code = code
        self._compiled = None

    def _compile(self):
        self._compiled = Template(self.code)

    def render(self, ctx):
        if not self._compiled:
            self._compile()
        return self._compiled.render(ctx)

SCOPE_TEMPLATES = defaultdict(lambda: SimpleTemplate('{{ url }}'), {
    'javascript': SimpleTemplate("<script src='{{ url }}?{{ checksum }}' {% if checksum %}integrity='sha256-{{ checksum }}'{% endif %} crossorigin='anonymous' ></script>"),
    'scripts': SimpleTemplate("<script src='{{ url }}?{{ checksum }} {% if checksum %}integrity='sha256-{{ checksum }}'{% endif %} crossorigin='anonymous' '></script>"),
    'js': SimpleTemplate("<script src='{{ url }}?{{ checksum }} {% if checksum %}integrity='sha256-{{ checksum }}'{% endif %} crossorigin='anonymous' '></script>"),

    'stylesheets': SimpleTemplate("<link rel='stylesheet' href='{{ url }}?{{ checksum }}' type='text/css' {% if checksum %}integrity='sha256-{{ checksum }}'{% endif %} crossorigin='anonymous' ></link>"),
    'styles': SimpleTemplate("<link rel='stylesheet' href='{{ url }}?{{ checksum }}' type='text/css' {% if checksum %}integrity='sha256-{{ checksum }}'{% endif %} crossorigin='anonymous' ></link>"),

    'images': SimpleTemplate("<img src='{{ url }}?{{ checksum }}'>"),
})


@register.simple_tag
@library.global_function
def asset(bundle_id):
    prefix, scope, name = Manifest.from_id(bundle_id)
    assets = Manifest.find_assets(bundle_id)
    if len(assets) == 0:
        raise Exception("Asset not found:", bundle_id)
    out = []
    for (url, data) in assets:
        ctx = Context(data)
        ctx['url'] = url
        tpl = SCOPE_TEMPLATES[scope]

        out.append(tpl.render(ctx))
    return mark_safe("\n".join(out))

@register.simple_tag
@library.global_function
def asset_url(bundle_id):
    assets = Manifest.find_assets(bundle_id)
    if len(assets) == 0:
        raise Exception("Asset not found:", bundle_id)
    return mark_safe(" ".join([url for url, _ in assets]))



from django.apps import apps
from django.utils.module_loading import import_string

class Filters(object):
    _app_filters_loaded = False
    FILTERS = {}


def register_filter(name):
    def filter_decorator(func):
        Filters.FILTERS[name] = func
        return func
    return filter_decorator


def get_filter_by_name(name):
    load_app_filters()
    if isinstance(name, str):
        return Filters.FILTERS[name]
    else:
        ret = Filters.FILTERS[name]
        setattr(ret, "_config", name.get('config',{}))
        return ret

def load_app_filters():
    if Filters._app_filters_loaded:
        return
    Filters._app_filters_loaded = True
    for app_cfg in apps.get_app_configs():
        try:
            __import__(app_cfg.name+'.assetfilters')
        except:
            pass

def get_all_filters():
    load_app_filters()
    return Filters.FILTERS.items()

from .cssmin import *
from .sass import *
from .clojure import *
from .ttf2woff2 import *
from .babel import *




import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890'

INTERNAL_IPS = ['127.0.0.1']

LOGGING_CONFIG = None   # avoids spurious output in tests


# Application definition

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'assets',
]

MEDIA_URL = 'http://localhost:7321/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'tests', 'cdn','media')
STATIC_URL = 'http://localhost:7321/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'tests', 'cdn','static')

MIDDLEWARE_CLASSES = [
    'django.middleware.common.CommonMiddleware',
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'NAME': 'jinja2',
        'BACKEND': 'django_jinja.backend.Jinja2',
        'APP_DIRS': True,
        'DIRS': [os.path.join(BASE_DIR, 'tests', 'templates', 'jinja2')],
    },
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request',
            ],
        },
    },
]

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, 'tests', 'static', 'manifest'),
    os.path.join(BASE_DIR, 'tests', 'static', 'standard'),
]

# Cache and database

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
    'second': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
    },
}

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}



import gevent.lock
import gevent.monkey
import gevent.pywsgi
import gevent.queue
gevent.monkey.patch_all(thread=False)

from geventwebsocket import WebSocketError
from geventwebsocket.handler import WebSocketHandler

import json
import logging
import time

from . import bottle
from .utils import MessageBroker, IP_ANY

logger = logging.getLogger(__name__)

LIVE_PORT = 35729

class LiveServer:

    def _websocket_handler(self):
        logger.info("Registering client")
        q = self._broker.sub()
        try:
            wsock = bottle.request.environ.get('wsgi.websocket')
            while True:
                message = q.get()
                wsock.send(json.dumps(message))
        except WebSocketError:
            pass
        finally:
            self._broker.unsub(q)

    def __init__(self, ip=IP_ANY, port=LIVE_PORT):
        self._broker = MessageBroker()
        self._app = bottle.Bottle()
        self._ip = ip
        self._port = port
        self._app.route('/')(self._websocket_handler)

    def serve_forever(self):
        logger.info("Starting livereload server on {ip}:{port}".format(ip=self._ip, port=self._port))
        self._server = gevent.pywsgi.WSGIServer((self._ip, self._port), self._app, handler_class=WebSocketHandler)
        self._server.serve_forever()

    def reload(self, path=None):
        self._broker.pub({'command': 'reload', 'path': path})

    def status(self, status):
        self._broker.pub({'command': 'status', 'status': status})

    @property
    def url(self):
        return 'ws://{ip}:{port}/'.format(ip=self._ip,port=self._port)







import logging

from django.core.management.base import BaseCommand, CommandError
from assets.manifest  import Manifest

FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Find asset manifests'

    def handle(self, *args, **options):
        Manifest.load()
        print(Manifest.ASSET_LIST)







from . import register_filter

try:
    import os
    from pathlib import Path
    from scss.compiler import Compiler

    @register_filter('sass')
    def sass_filter(bytes, cwd, config={}):
        """
            A sass compiler based on the Python scss module
        """
        compiler = Compiler(search_path=(cwd,))
        return compiler.compile_string(bytes)
except:
    from .utils import pipe

    @register_filter('sass')
    def sass_filter(bytes, cwd, config={}):
        """
            A sass compiler based on the sass executable
        """
        return pipe('sass --scss', bytes, cwd=cwd)



import subprocess


def pipe(program, input, cwd=None):
    """
        Pipe :param:`input` into the program :param:`program` run in
        the directory :param:`cwd`. Returns the program output stream.
    """
    p = subprocess.Popen(program, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.DEVNULL, shell=True, cwd=cwd, bufsize=-1)
    return p.communicate(bytes(input, encoding='utf-8'))[0]



import gevent
import gevent.lock
import gevent.monkey
import gevent.queue
gevent.monkey.patch_all(thread=False)

import logging
import os
import signal

logger = logging.getLogger(__name__)

IP_ANY = '0.0.0.0'

class MessageBroker:
    def __init__(self):
        self.q_lock = gevent.lock.BoundedSemaphore(1)
        self.queues = []

    def sub(self):
        queue = gevent.queue.Queue()
        self.q_lock.acquire()
        self.queues.append(queue)
        self.q_lock.release()
        return queue

    def unsub(self, queue, exit_on_empty=None):
        self.q_lock.acquire()
        self.queues.remove(queue)
        self.q_lock.release()
        if exit_on_empty is not None:
            if len(self.queues) == 0:
                exit(exit_on_empty)

    def pub(self, message):
        self.q_lock.acquire()
        queues = [q for q in self.queues]
        self.q_lock.release()
        logger.info("Sending message '"+str(message)+"' to "+str(len(queues))+" client(s).")
        for q in queues:
            q.put(message)

    def __len__(self):
        self.q_lock.acquire()
        ret = len(self.queues)
        self.q_lock.release()
        return ret

class CleanChildProcesses:
    """
        Taken from https://stackoverflow.com/questions/320232/ensuring-subprocesses-are-dead-on-exiting-python-program
    """
    def __enter__(self):
        os.setpgrp() # create new process group, become its leader
    def __exit__(self, type, value, traceback):
        try:
            os.killpg(0, signal.SIGINT) # kill all processes in my group
        except KeyboardInterrupt:
            # SIGINT is delievered to this process as well as the child processes.
            # Ignore it so that the existing exception, if any, is returned. This
            # leaves us with a clean exit code if there was no exception.
            pass

def with_lock(lock, func):
    def wrapped(*args, **kwargs):
        lock.acquire()
        try:
           return func(*args, **kwargs)
        finally:
            lock.release()
    return wrapped

def pid_exists(pid):
    """
        Check For the existence of a unix pid.

        https://stackoverflow.com/questions/568271/how-to-check-if-there-exists-a-process-with-a-given-pid-in-python
    """
    if pid is None:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True



from . import register_filter
from .utils import pipe

try:
    import shutil

    if shutil.which('ttf2woff2') is not None:
        @register_filter('ttf2woff2')
        def ttf2woff2_filter(bytes, cwd, config={}):
            """
            A filter for converting ttf to woff2 fonts (based on google's woff2 library), installable via
            npm -g install ttf2woff2
            """
            return pipe('ttf2woff2', bytes, cwd=cwd)
    else:
        @register_filter('ttf2woff2')
        def ttf2woff2_filter(bytes, cwd, config={}):
            """
            A stub for ttf2woff2 converter (the stub does nothing)
            """
            return bytes
except:
    @register_filter('ttf2woff2')
    def ttf2woff2_filter(bytes, cwd, config={}):
        """
        A stub for ttf2woff2 converter (the stub does nothing)
        """
        return bytes


